"""
This script defines object-relational mapping helpers to ease
 the manipulation of Kili data structures.
"""
from dataclasses import dataclass

class DictClass(dict):
    """
    A python class that acts like dict
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self

@dataclass
class AnnotationFormat:
    # pylint: disable=invalid-name
    """
    List of annotation formats
    """

    Latest = 'latest'
    Raw = 'raw'
    Simple = 'simple'
    YoloV4Pytorch = 'yolo_v4_pytorch'


@dataclass
class JobMLTask:
    # pylint: disable=invalid-name
    """
    List of ML Tasks
    """
    AssetAnnotation = 'ASSET_ANNOTATION'
    Classification = 'CLASSIFICATION'
    NamedEntitiesRecognition = 'NAMED_ENTITIES_RECOGNITION'
    NamedEntitiesRelation = 'NAMED_ENTITIES_RELATION'
    ObjectRelation = 'OBJECT_RELATION'
    ObjectDetection = 'OBJECT_DETECTION'
    PoseEstimation = 'POSE_ESTIMATION'
    RegionDetection = 'REGION_DETECTION'
    SpeechToText = 'SPEECH_TO_TEXT'
    Transcription = 'TRANSCRIPTION'

@dataclass
class JobTool:
    # pylint: disable=invalid-name
    """
    List of tools
    """
    Marker = 'marker'
    Polygon = 'polygon'
    Polyline = 'polyline'
    Range = 'range'
    Rectangle = 'rectangle'
    Semantic = 'semantic'
    Vector = 'vector'


def get_polygon(annotation):
    """
    Extracts a polygon from an annotation

    Parameters
    ----------
    - annotation : Kili annotation
    """
    try:
        return annotation['boundingPoly'][0]['normalizedVertices']
    except KeyError:
        return None


def get_category(annotation):
    """
    Extracts a category from an annotation

    Parameters
    ----------
    - annotation : Kili annotation
    """
    try:
        return annotation['categories'][0]['name']
    except KeyError:
        return None


def get_named_entity(annotation):
    """
    Extracts a named entity from an annotation

    Parameters
    ----------
    - annotation : Kili annotation
    """
    try:
        return {
            'beginId': annotation['beginId'],
            'beginOffset': annotation['beginOffset'],
            'content': annotation['content'],
            'endId': annotation['endId'],
            'endOffset': annotation['endOffset']
        }
    except KeyError:
        return None


def format_image_annotation(annotation):
    """
    Extracts a category, polygon, named entity from an annotation
    depending of the annotation type

    Parameters
    ----------
    - annotation : Kili annotation
    """
    category = get_category(annotation)
    polygon = get_polygon(annotation)
    named_entity = get_named_entity(annotation)
    if polygon is not None:
        return (category, polygon)
    if named_entity is not None:
        return (category, named_entity)
    if category is not None:
        return category
    return None


class Label(DictClass):
    """
    Label class
    """

    jsonResponse = {}

    def json_response(self, _format=AnnotationFormat.Raw):
        """
        Format a json response

        Parameters
        ----------
        - _format: expected format
        """
        if 'jsonResponse' not in self:
            raise Exception(
                'You did not fetch jsonResponse for' \
                f' label "{self["id"] if "id" in self else self}"')
        if _format == AnnotationFormat.Raw:
            return self.jsonResponse
        if _format == AnnotationFormat.Simple:
            job_names = self.jsonResponse.keys()
            if len(job_names) > 1:
                return {
                    'error': 'Simple format is not adapted' \
                    ' when there is more than one job.' \
                    ' Please choose another annotation format.'}
            for job_name in job_names:
                job_response = self.jsonResponse[job_name]
                category = get_category(job_response)
                if category is not None:
                    return category
                if 'annotations' not in job_response:
                    return None
                return [format_image_annotation(annotation) \
                    for annotation in job_response['annotations']]
            return None
        raise Exception(f'format "{_format}" is not a valid annotation format.')


class Asset(DictClass):
    """
    Asset class
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if 'labels' in self:
            labels = []
            for label in self['labels']:
                labels.append(Label(label))
            self.labels = labels
        if 'latestLabel' in self:
            latest_label = self['latestLabel']
            if latest_label is not None:
                self.latestLabel = Label(latest_label) # pylint: disable=invalid-name



"""
Project queries
"""

from typing import Optional
from typeguard import typechecked

from ...helpers import Compatible, deprecate, format_result, fragment_builder
from .queries import gql_projects, GQL_PROJECTS_COUNT
from ...types import Project
from ...constants import NO_ACCESS_RIGHT


class QueriesProject:
    """
    Set of Project queries
    """
    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """
        Initializes the subclass

        Parameters
        ----------
        - auth : KiliAuth object
        """
        self.auth = auth

    # pylint: disable=dangerous-default-value
    @Compatible(['v1', 'v2'])
    @typechecked
    def projects(self,
                 project_id: Optional[str] = None,
                 search_query: Optional[str] = None,
                 should_relaunch_kpi_computation: Optional[bool] = None,
                 updated_at_gte: Optional[str] = None,
                 updated_at_lte: Optional[str] = None,
                 skip: int = 0,
                 fields: list = [
                     'consensusTotCoverage',
                     'id',
                     'inputType',
                     'interfaceCategory',
                     'jsonInterface',
                     'maxWorkerCount',
                     'minAgreement',
                     'minConsensusSize',
                     'reviewCoverage',
                     'roles.id',
                     'roles.role',
                     'roles.user.email',
                     'roles.user.id',
                     'roles.user.name',
                     'title'],
                 first: int = 100):
        # pylint: disable=line-too-long
        """
        Get projects given a set of criteria

        Parameters
        ----------
        - project_id : str, optional (default = None)
            Select a specific project through its project_id.
        - search_query : str, optional (default = None)
            Returned projects with a title or a description matching this string.
        - should_relaunch_kpi_computation : bool, optional (default = None)
            Technical field, added to indicate changes in honeypot or consensus settings.
        - updated_at_gte : string, optional (default = None)
            Returned projects should have a label whose update date is greated or equal
            to this date.
            Formatted string should have format : "YYYY-MM-DD"
        - updated_at_lte : string, optional (default = None)
            Returned projects should have a label whose update date is lower or equal to this date.
            Formatted string should have format : "YYYY-MM-DD"
        - skip : int, optional (default = 0)
            Number of projects to skip (they are ordered by their creation).
        - fields : list of string, optional (default = ['consensusTotCoverage', 'id',
            'inputType', 'interfaceCategory', 'jsonInterface', 'maxWorkerCount', 'minAgreement',
            'minConsensusSize', 'roles.id', 'roles.role', 'roles.user.email', 'roles.user.id',
            'roles.user.name', 'title'])
            All the fields to request among the possible fields for the projects.
            See [the documentation](https://cloud.kili-technology.com/docs/python-graphql-api/graphql-api/#project) for all possible fields.
        - first : int , optional (default = 100)
            Maximum number of projects to return. Can only be between 0 and 100.

        Returns
        -------
        - a result object which contains the query if it was successful, or an error message else.

        Examples
        -------
        >>> # List all my projects
        >>> kili.projects()
        """
        _gql_projects = gql_projects(fragment_builder(fields, Project))
        variables = {
            'where': {
                'id': project_id,
                'searchQuery': search_query,
                'shouldRelaunchKpiComputation': should_relaunch_kpi_computation,
                'updatedAtGte': updated_at_gte,
                'updatedAtLte': updated_at_lte,
            },
            'skip': skip,
            'first': first
        }
        result = self.auth.client.execute(_gql_projects, variables)
        return format_result('data', result)

    @Compatible(['v1', 'v2'])
    @typechecked
    def count_projects(
            self,
            project_id: Optional[str] = None,
            search_query: Optional[str] = None,
            should_relaunch_kpi_computation: Optional[bool] = None,
            updated_at_gte: Optional[str] = None,
            updated_at_lte: Optional[str] = None):
        """
        Counts the number of projects with a search_query

        Parameters
        ----------
        - project_id : str, optional (default = None)
            Select a specific project through its project_id
        - search_query : str, optional (default = None)
            Returned projects have a title or a description that matches this string.
        - should_relaunch_kpi_computation : bool, optional (default = None)
            Technical field, added to indicate changes in honeypot or consensus settings
        - updated_at_gte : string, optional (default = None)
            Returned projects should have a label whose update date is greated
            or equal to this date.
            Formatted string should have format : "YYYY-MM-DD"
        - updated_at_lte : string, optional (default = None)
            Returned projects should have a label whose update date is lower or equal to this date.
            Formatted string should have format : "YYYY-MM-DD"

        Returns
        -------
        - a positive integer corresponding to the number of results of the query
            if it was successful, or an error message else.
        """
        variables = {
            'where': {
                'id': project_id,
                'searchQuery': search_query,
                'shouldRelaunchKpiComputation': should_relaunch_kpi_computation,
                'updatedAtGte': updated_at_gte,
                'updatedAtLte': updated_at_lte,
            }
        }
        result = self.auth.client.execute(GQL_PROJECTS_COUNT, variables)
        count = format_result('data', result)
        return count



"""
Project user queries
"""

from typing import Optional

from typeguard import typechecked

from ...helpers import Compatible, deprecate, format_result, fragment_builder
from .queries import gql_project_users, GQL_PROJECT_USERS_COUNT
from ...types import ProjectUser


class QueriesProjectUser:
    """
    Set of ProjectUser queries
    """
    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """
        Initializes the subclass

        Parameters
        ----------
        - auth : KiliAuth object
        """
        self.auth = auth

    # pylint: disable=dangerous-default-value,invalid-name
    @Compatible(['v1', 'v2'])
    @typechecked
    def project_users(self,
                      email: Optional[str] = None,
                      id: Optional[str] = None, # pylint: disable=redefined-builtin
                      organization_id: Optional[str] = None,
                      project_id: Optional[str] = None,
                      fields: list = ['activated', 'id', 'role',
                                      'starred', 'user.email', 'user.id', 'user.name'],
                      first: int = 100,
                      skip: int = 0):
        # pylint: disable=line-too-long
        """
        Return projects and their users (possibly with their KPIs) respecting a set of criteria

        Parameters
        ----------
        - email : str, optional (default = None)
        - organization_id : str, optional (default = None)
        - project_id : str, optional (default = None)
        - fields : list, optional (default = ['activated', 'id', 'role', 'starred',
            'user.email', 'user.id', 'user.name'])
            All the fields to request among the possible fields for the projectUsers.
            See [the documentation](https://cloud.kili-technology.com/docs/python-graphql-api/graphql-api/#projectuser) for all possible fields.
        - first : int, optional (default = 100)
            Maximum number of users to return. Can only be between 0 and 100.
        - skip : int, optional (default = 0)
            Number of project users to skip

        Returns
        -------
        - a result object which contains the query if it was successful, or an error message else.

        Examples
        -------
        >>> # Retrieve consensus marks of all users in project
        >>> kili.project_users(project_id=project_id, fields=['consensusMark', 'user.email'])
        """
        variables = {
            'first': first,
            'skip': skip,
            'where': {
                'id': id,
                'project': {
                    'id': project_id,
                },
                'user': {
                    'email': email,
                    'organization': {
                        'id': organization_id,
                    }
                },
            }
        }
        _gql_project_users = gql_project_users(
            fragment_builder(fields, ProjectUser))
        result = self.auth.client.execute(_gql_project_users, variables)
        return format_result('data', result)

    # pylint: disable=invalid-name
    @typechecked
    def count_project_users(
            self,
            email: Optional[str] = None,
            id: Optional[str] = None, # pylint: disable=redefined-builtin
            organization_id: Optional[str] = None,
            project_id: Optional[str] = None):
        """
        Counts the number of projects and their users respecting a set of criteria

        Parameters
        ----------
        - email : str, optional (default = None)
        - organization_id : str, optional (default = None)
        - project_id : str, optional (default = None)

        Returns
        -------
        - a positive integer corresponding to the number of results of the query
            if it was successful, or an error message else.
        """
        variables = {
            'where': {
                'id': id,
                'project': {
                    'id': project_id,
                },
                'user': {
                    'email': email,
                    'organization': {
                        'id': organization_id,
                    }
                },
            }
        }
        result = self.auth.client.execute(GQL_PROJECT_USERS_COUNT, variables)
        count = format_result('data', result)
        return count



"""
Issue queries
"""

from dataclasses import dataclass
from typing import List, Optional

from typeguard import typechecked

from ...helpers import Compatible, format_result, fragment_builder
from .queries import gql_issues
from ...types import Issue as IssueType

@dataclass
class QueriesIssue:
    """
    Set of Issue queries
    """
    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """
        Initializes the subclass

        Parameters
        ----------
        - auth : KiliAuth object
        """
        self.auth = auth

    # pylint: disable=dangerous-default-value
    @Compatible(['v1', 'v2'])
    @typechecked
    def issues(self,
               fields: Optional[list] = [
                   'id',
                   'createdAt',
                   'hasBeenSeen',
                   'issueNumber',
                   'status',
                   'type'],
               first: Optional[int] = 100,
               project_id: Optional[str] = None,
               skip: Optional[int] = 0):
        """
        Get an array of issues given a set of criteria

        Parameters
        ----------
        - project_id : str, optional (default = None)
            Project ID the issue belongs to.

        Returns
        -------
        - a result object which contains the query if it was successful, or an error message else.

        Examples
        -------
        >>> # List all issues of a project and their authors
        >>> kili.issues(project_id=project_id, fields=['author.email'])
        """
        variables = {
            'where': {
                'project': {
                    'id': project_id,
                },
            },
            'skip': skip,
            'first': first,
        }
        _gql_issues = gql_issues(fragment_builder(fields, IssueType))
        result = self.auth.client.execute(_gql_issues, variables)
        return format_result('data', result)



"""
This script permits to initialize the Kili Playground client.
"""
import os

from kili.mutations.api_key import MutationsApiKey
from kili.mutations.asset import MutationsAsset
from kili.mutations.dataset import MutationsDataset
from kili.mutations.dataset_asset import MutationsDatasetAsset
from kili.mutations.label import MutationsLabel
from kili.mutations.notification import MutationsNotification
from kili.mutations.organization import MutationsOrganization
from kili.mutations.project import MutationsProject
from kili.mutations.project_version import MutationsProjectVersion
from kili.mutations.user import MutationsUser
from kili.queries.asset import QueriesAsset
from kili.queries.dataset_asset import QueriesDatasetAsset
from kili.queries.issue import QueriesIssue
from kili.queries.label import QueriesLabel
from kili.queries.lock import QueriesLock
from kili.queries.organization import QueriesOrganization
from kili.queries.notification import QueriesNotification
from kili.queries.project import QueriesProject
from kili.queries.project_user import QueriesProjectUser
from kili.queries.project_version import QueriesProjectVersion
from kili.queries.user import QueriesUser
from kili.subscriptions.label import SubscriptionsLabel


from kili.authentication import KiliAuth


class Kili(  # pylint: disable=too-many-ancestors
        MutationsApiKey,
        MutationsAsset,
        MutationsDataset,
        MutationsDatasetAsset,
        MutationsLabel,
        MutationsNotification,
        MutationsOrganization,
        MutationsProject,
        MutationsProjectVersion,
        MutationsUser,
        QueriesAsset,
        QueriesDatasetAsset,
        QueriesIssue,
        QueriesLabel,
        QueriesLock,
        QueriesOrganization,
        QueriesNotification,
        QueriesProject,
        QueriesProjectUser,
        QueriesProjectVersion,
        QueriesUser,
        SubscriptionsLabel):
    """
    Kili Client.
    """

    def __init__(self, api_key=os.getenv('KILI_API_KEY'),
                 api_endpoint='https://cloud.kili-technology.com/api/label/v2/graphql',
                 verify=True):
        """
        Parameters
        ----------
        api_key:
            User API key generated from https://cloud.kili-technology.com/label/my-account/api-key

        api_endpoint : str
            Recipient of the HTTP operation

        verify : bool
            Verify certificate. Set to False on local deployment without SSL.

        Returns
        -------
        Kili object
        Object container your API session.
        Then, list:
        - your assets with: kili.assets()
        - your labels with: kili.labels()
        - your projects with: kili.projects()
        """
        self.auth = KiliAuth(
            api_key=api_key, api_endpoint=api_endpoint, verify=verify)
        super().__init__(self.auth)



"""
Project version mutations
"""

from dataclasses import dataclass
from typing import Optional

from typeguard import typechecked

from ...helpers import Compatible, format_result
from .queries import (GQL_UPDATE_PROPERTIES_IN_PROJECT_VERSION)

@dataclass
class MutationsProjectVersion:
    """
    Set of ProjectVersion mutations
    """
    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """
        Initializes the subclass

        Parameters
        ----------
        - auth : KiliAuth object
        """
        self.auth = auth

    @Compatible(['v2'])
    @typechecked
    def update_properties_in_project_version(
            self,
            project_version_id: str,
            content: Optional[str]):
        """
        Update properties of a project version

        Parameters
        ----------
        - project_version_id : str
            Identifier of the project version
        - content : Optional[str]
            Link to download the project version

        Returns
        -------
        - a result object which indicates if the mutation was successful.

        Examples
        -------
        >>> kili.update_properties_in_project_version(
                project_version_id=project_version_id,
                content='test')
        """
        variables = {
            'content': content,
            'id': project_version_id,
        }
        result = self.auth.client.execute(
            GQL_UPDATE_PROPERTIES_IN_PROJECT_VERSION, variables)
        return format_result('data', result)



"""
Queries of dataset mutations
"""

from .fragments import DATASET_FRAGMENT

GQL_APPEND_TO_DATASET = f'''
mutation(
    $data: AppendToDatasetData!
    $where: DatasetWhere!
) {{
  data: appendToDataset(
    data: $data,
    where: $where
  ) {{
    {DATASET_FRAGMENT}
  }}
}}
'''

GQL_APPEND_TO_DATASET_USERS = f'''
mutation(
    $data: AppendToDatasetUsersData!
    $where: DatasetWhere!
) {{
  data: appendToDatasetUsers(
    data: $data,
    where: $where
  ) {{
    {DATASET_FRAGMENT}
  }}
}}
'''

GQL_APPEND_DATASETS_TO_PROJECT = '''
mutation(
    $data: AppendDatasetsToProjectData!
    $where: ProjectWhere!
) {
  data: appendDatasetsToProject(
    data: $data,
    where: $where
  )
}
'''

GQL_CREATE_DATASET = f'''
mutation(
    $data: CreateDatasetData!
) {{
  data: createDataset(
    data: $data,
  ) {{
    {DATASET_FRAGMENT}
  }}
}}
'''

GQL_DELETE_DATASET = f'''
mutation(
    $where: DatasetWhere!
) {{
  data: deleteDataset(
    where: $where
  ) {{
    {DATASET_FRAGMENT}
  }}
}}
'''

GQL_UPDATE_PROPERTIES_IN_DATASET = f'''
mutation(
    $datasetId: ID!
    $assetType: InputType!
    $name: String!
) {{
  data: updatePropertiesInDataset(
    data: {{
      assetType: $assetType,
      name: $name
    }}
    where: {{
      id: $datasetId
    }}) {{
    {DATASET_FRAGMENT}
  }}
}}
'''



"""
Api key mutations
"""

from typeguard import typechecked
from ...helpers import Compatible, deprecate, format_result

from .queries import GQL_APPEND_TO_API_KEYS

class MutationsApiKey: # pylint: disable=too-few-public-methods
    """
    Set of User mutations
    """
    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """
        Initializes the subclass

        Parameters
        ----------
        - auth : KiliAuth object
        """
        self.auth = auth

    @Compatible(['v1', 'v2'])
    @typechecked
    def append_to_api_keys(self, api_key: str, name: str):
        """
        Create an api key to connect to the API

        Parameters
        ----------
        - api_key : str, a new api key to connect with
        - name : str, a name used to describe the api key.
        """
        variables = {
            'data': {'key': api_key,
                     'name': name},
            'where': {'email': self.auth.user_email}
        }
        result = self.auth.client.execute(GQL_APPEND_TO_API_KEYS, variables)
        return format_result('data', result)



"""
Helpers for GraphQL Queries and Mutations
"""

import base64
import functools
from json import dumps, loads
import re
import warnings

import requests

class Compatible():
    """
    Compatibility of Kili playground version with Kili API version
    """

    # pylint: disable=dangerous-default-value
    def __init__(self, endpoints=['v1']):
        self.endpoints = endpoints
        self.version_extractor = re.compile(r'\/v\d+/')
        self.address_extractor = re.compile(r':400\d+/')

    def client_is_compatible(self, endpoint: str):
        """
        Checks if client is compatible with Kili API

        Parameters
        ----------
        - endpoint: the Kili API endpoint
        """
        version_matched = self.version_extractor.search(endpoint)
        address_matched = self.address_extractor.search(endpoint)
        if not version_matched and not address_matched:
            return False
        if address_matched:
            version = 'v1' if address_matched.group() == ':4000/' else 'v2'
        if version_matched:
            version = 'v1' if version_matched.group() == '/v1/' else 'v2'
        return version in self.endpoints

    def __call__(self, resolver, *args, **kwargs):
        @functools.wraps(resolver)
        def checked_resolver(*args, **kwargs):
            try:
                client_endpoint = args[0].auth.client.endpoint
            except Exception as exception:
                raise ValueError(
                    'Cannot find client endpoint from resolver' \
                    f' {resolver.__name__} with arguments {args}') from exception
            if self.client_is_compatible(client_endpoint):
                return resolver(*args, **kwargs)
            raise EndpointCompatibilityError(
                resolver.__name__, client_endpoint)
        return checked_resolver


class EndpointCompatibilityError(Exception):
    """
    EndpointCompatibilityError
    """
    def __init__(self, resolver, endpoint):
        super().__init__(
            f'Resolver {resolver} is not compatible with the following endpoint : {endpoint}')


class GraphQLError(Exception):
    """
    GraphQLError
    """
    def __init__(self, mutation, error):
        super().__init__(f'Mutation "{mutation}" failed with error: "{error}"')


def format_result(name, result, _object=None):
    """
    Formats the result of the GraphQL queries.

    Parameters
    ----------
    - name: name of the field to extract, usually data
    - result: query result to parse
    """
    if 'errors' in result:
        raise GraphQLError(name, result['errors'])
    formatted_json = format_json(result['data'][name])
    if _object is None:
        return formatted_json
    if isinstance(formatted_json, list):
        return [_object(element) for element in formatted_json]
    return _object(formatted_json)


def content_escape(content):
    """
    Escapes the content

    Parameters
    ----------
    - content: string to escape
    """
    return content.replace('\\', '\\\\').replace('\n', '\\n').replace('"', '\\"')


def get_data_type(path):
    """
    Get the data type, either image/png or application/pdf

    Parameters
    ----------
    - path: path of the file
    """
    if re.match(r'.*(jpg|jpeg)$', path.lower()):
        return 'image/png'
    if re.match(r'.*pdf$', path.lower()):
        return 'application/pdf'
    return ''


def encode_base64(path):
    """
    Encode a file in base 64

    Parameters
    ----------
    - path: path of the file
    """
    data_type = get_data_type(path)
    with open(path, 'rb') as image_file:
        return f'data:{data_type};base64,' + \
            base64.b64encode(image_file.read()).decode('ascii')


def is_url(path):
    """
    Check if the path is a url or something else

    Parameters
    ----------
    - path: path of the file
    """
    return isinstance(path, str) and re.match(r'^(http://|https://)', path.lower())


def format_json_dict(result):
    """
    Formats the dict part of a json return by a GraphQL query into a python object

    Parameters
    ----------
    - result: result of a GraphQL query
    """
    for key, value in result.items():
        if key in ['jsonInterface', 'jsonMetadata', 'jsonResponse', 'rules']:
            if (value == '' or value is None) \
                    and not (is_url(value) and key == 'jsonInterface'):
                result[key] = {}
            elif isinstance(value, str):
                try:
                    if is_url(value):
                        result[key] = requests.get(value).json()
                    else:
                        result[key] = loads(value)
                except Exception as exception:
                    raise ValueError(
                        'Json Metadata / json response /' \
                        ' json interface / rules should be valid jsons') from exception
        else:
            result[key] = format_json(value)
    return result


def format_json(result):
    """
    Formats the json return by a GraphQL query into a python object

    Parameters
    ----------
    - result: result of a GraphQL query
    """
    if result is None:
        return result
    if isinstance(result, list):
        return [format_json(elem) for elem in result]
    if isinstance(result, dict):
        return format_json_dict(result)
    return result


def fragment_builder(fields, type_of_fields):
    """
    Builds a GraphQL fragment for a list of fields to query

    Parameters
    ----------
    - fields
    - type_of_fields
    """
    fragment = ''
    subfields = [field.split('.', 1) for field in fields if '.' in field]
    if subfields:
        for subquery in {subfield[0] for subfield in subfields}:
            type_of_fields_subquery = getattr(type_of_fields, subquery)
            try:
                if issubclass(type_of_fields_subquery, object):
                    fields_subquery = [subfield[1]
                                       for subfield in subfields if subfield[0] == subquery]
                    new_fragment = fragment_builder(fields_subquery, type_of_fields_subquery)
                    fragment += f' {subquery}{{{new_fragment}}}'
            except ValueError:
                print(f'{subquery} must be a valid subquery field')
        fields = [field for field in fields if '.' not in field]
    for field in fields:
        try:
            getattr(type_of_fields, field)
        except ValueError:
            print(f'{field} must be an instance of {type_of_fields}')
        if isinstance(field, str):
            fragment += f' {field}'
        else:
            raise Exception('Please provide the fields to query as strings')
    return fragment


def deprecate(msg, _type=DeprecationWarning):
    """
    Decorator that generates a deprecation message

    Parameters
    ----------
    - msg: string message
    - type: DeprecationWarning by default
    """
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            warnings.warn(msg, _type, stacklevel=2)
            return func(*args, **kwargs)
        return wrapper
    return decorator


def format_metadata(metadata):
    """
    Formats metadata

    Parameters
    ----------
    - metadata: a python object
    """
    if metadata is None:
        return metadata
    if isinstance(metadata, str):
        return metadata
    if isinstance(metadata, (dict, list)):
        return dumps(metadata)
    raise Exception(
        f'Metadata {metadata} of type {type(metadata)} must either be None,' \
        ' a string a list or a dict.')


def convert_to_list_of_none(array, length):
    """
    Turns a value in a list of length length

    Parameters
    ----------
    - array
    - length
    """
    if isinstance(array, list):
        if len(array) != length:
            raise Exception(f'array should have length {length}')
        return array
    return [None] * length


def is_none_or_empty(_object):
    """
    Tests if an object is none or empty

    Parameters
    ----------
    - object: a python object
    """
    object_is_empty = isinstance(_object, list) and len(_object) == 0
    return _object is None or object_is_empty


def list_is_not_none_else_none(_object):
    """
    Formats an object as a singleton if not none

    Parameters
    ----------
    - object: a python object
    """
    return [_object] if _object is not None else None


def infer_id_from_external_id(playground, asset_id: str, external_id: str, project_id: str):
    """
    Infer asset id from external id

    Parameters
    ----------
    - asset_id: asset id
    - external_id: external id
    - project_id: project id
    """
    if asset_id is None and external_id is None:
        raise Exception(
            'Either provide asset_id or external_id and project_id')
    if asset_id is not None:
        return asset_id
    assets = playground.assets(
        external_id_contains=[external_id], project_id=project_id, fields=['id'], disable_tqdm=True)
    if len(assets) == 0:
        raise Exception(
            f'No asset found with external ID "{external_id}"')
    if len(assets) > 1:
        raise Exception(
            f'Several assets found containing external ID "{external_id}":' \
            f' {assets}. Please, use asset ID instead.')
    return assets[0]['id']



"""
Queries of dataset asset mutations
"""

from .fragments import ASSET_FRAGMENT


GQL_DELETE_DATASET_ASSETS = '''
mutation(
    $where: DatasetAssetWhere!
) {
  data: deleteDatasetAssets(
    where: $where
  )
}
'''

GQL_UPDATE_PROPERTIES_IN_DATASET_ASSETS = f'''
mutation(
    $whereArray: [DatasetAssetWhere!]!
    $dataArray: [DatasetAssetData!]!
) {{
  data: updatePropertiesInDatasetAssets(
    where: $whereArray,
    data: $dataArray
  ) {{
    {ASSET_FRAGMENT}
  }}
}}
'''



"""
Queries of project version mutations
"""

from .fragments import PROJECT_VERSION_FRAGMENT

GQL_UPDATE_PROPERTIES_IN_PROJECT_VERSION = f'''
mutation(
    $id: ID!
    $content: String
) {{
  data: updatePropertiesInProjectVersion(
    where: {{id: $id}}
    data: {{
      content: $content
    }}
  ) {{
    {PROJECT_VERSION_FRAGMENT}
  }}
}}
'''



"""
Dataset asset mutations
"""

from dataclasses import dataclass
from typing import List, Optional
from functools import partial

from typeguard import typechecked

from ...helpers import (Compatible,
                        convert_to_list_of_none,
                        format_metadata,
                        format_result)
from .queries import GQL_DELETE_DATASET_ASSETS, GQL_UPDATE_PROPERTIES_IN_DATASET_ASSETS
from ...orm import Asset

@dataclass
class MutationsDatasetAsset:
    """
    Set of DatasetAsset mutations
    """
    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """
        Initializes the subclass

        Parameters
        ----------
        - auth : KiliAuth object
        """
        self.auth = auth

    @Compatible(['v2'])
    @typechecked
    def delete_dataset_assets(self, asset_ids: List[str]):
        """
        Deletes assets of a dataset.

        Parameters
        ----------
        - asset_ids : List[str]
            Ids of the assets to delete

        Returns
        -------
        - a result object which indicates if the mutation was successful, or an error message else.

        Examples
        -------
        kili.delete_dataset_assets(
            asset_ids=['ckg22d81r0jrg0885unmuswj8'])
        """

        variables = {
            'where': {'idIn': asset_ids},
        }
        result = self.auth.client.execute(GQL_DELETE_DATASET_ASSETS, variables)
        return format_result('data', result)

    @Compatible(['v2'])
    @typechecked
    def update_properties_in_dataset_assets(self, asset_ids: List[str],
                                            external_ids: Optional[List[str]] = None,
                                            json_metadatas: Optional[List[dict]] = None,
                                            contents: Optional[List[str]
                                                               ] = None,
                                            json_contents: Optional[List[str]] = None):
        """
        Update the properties of one or more assets of datasets.

        Parameters
        ----------
        - asset_ids : List[str]
            The asset IDs to modify
        - external_ids : List[str], optional (default = None)
            Change the external id of the assets
        - json_metadatas : List[dict] , optional (default = None)
            The metadata given to an asset should be stored in a json like dict with keys
            "imageUrl", "text", "url".
            json_metadata = {'imageUrl': '','text': '','url': ''}
        - contents : List[str] (default = None)
            - For a NLP project, the content can be directly in text format
            - For an Image / Video / Pdf project, the content must be hosted on a web server,
            and you point Kili to your data by giving the URLs
        - json_contents : List[str] (default = None)
            - For a NLP project, the json_content is a a text formatted using RichText
            - For a Video project, the json_content is a json containg urls pointing to
                each frame of the video.

        Returns
        -------
        - a result object which indicates if the mutation was successful, or an error message else.

        Examples
        -------
        kili.update_properties_in_dataset_assets(
                asset_ids=["ckg22d81r0jrg0885unmuswj8", "ckg22d81s0jrh0885pdxfd03n"],
                contents=[None, 'https://to/second/asset.png'],
                external_ids=['external-id-of-your-choice-1', 'external-id-of-your-choice-2']
        )
        """

        formatted_json_metadatas = None
        if json_metadatas is None:
            formatted_json_metadatas = None
        else:
            if isinstance(json_metadatas, list):
                formatted_json_metadatas = list(
                    map(format_metadata, json_metadatas))
            else:
                raise Exception('json_metadatas',
                                'Should be either a None or a list of None, string, list or dict')

        where_array = [{'id': asset_id} for asset_id in asset_ids]
        nb_assets_to_modify = len(where_array)
        if nb_assets_to_modify > 100:
            raise Exception(
                f'Too many assets ({nb_assets_to_modify}) updated at a time')
        data_array = [{} for i in range(len(where_array))]
        list_of_properties = [external_ids,
                              formatted_json_metadatas, contents, json_contents]
        data = list(map(partial(convert_to_list_of_none,
                                length=nb_assets_to_modify), list_of_properties))
        property_names = [
            'externalId',
            'jsonMetadata',
            'content',
            'jsonContent',
            'status'
        ]
        for i, properties in enumerate(zip(*data)):
            for _property, property_value in zip(property_names, properties):
                data_array[i][_property] = property_value
        variables = {
            'whereArray': where_array,
            'dataArray': data_array
        }
        result = self.auth.client.execute(
            GQL_UPDATE_PROPERTIES_IN_DATASET_ASSETS, variables)
        return format_result('data', result, Asset)



"""
Organization mutations
"""

import json
from typing import Optional

from typeguard import typechecked

from ...helpers import Compatible, format_result
from .queries import (GQL_CREATE_ORGANIZATION,
                      GQL_UPDATE_PROPERTIES_IN_ORGANIZATION)


class MutationsOrganization:
    """
    Set of Organization mutations
    """
    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """
        Initializes the subclass

        Parameters
        ----------
        - auth : KiliAuth object
        """
        self.auth = auth

    @Compatible(['v1', 'v2'])
    @typechecked
    def create_organization(self, name: str, address: str, zip_code: str, city: str, country: str):
        """
        Create an organization

        Each user must be linked to an organization

        Parameters
        ----------
        - name : str
        - address : str
        - zip_code : str
        - city : str
        - country : str

        Returns
        -------
        - a result object which indicates if the mutation was successful, or an error message else.
        """
        variables = {
            'data': {'name': name,
                     'address': address,
                     'zipCode': zip_code,
                     'city': city,
                     'country': country}
        }
        result = self.auth.client.execute(GQL_CREATE_ORGANIZATION, variables)
        return format_result('data', result)

    @Compatible(['v1', 'v2'])
    @typechecked
    def update_properties_in_organization(self,
                                          organization_id: str,
                                          name: Optional[str] = None,
                                          address: Optional[str] = None,
                                          zip_code: Optional[str] = None,
                                          city: Optional[str] = None,
                                          country: Optional[str] = None,
                                          license: Optional[dict] = None):  # pylint: disable=redefined-builtin
        """
        Modify an organization

        Parameters
        ----------
        - organization_id : str
        - name : str
        - address : str
        - license : dict
        - zip_code : str
        - city : str
        - country : str

        Returns
        -------
        - a result object which indicates if the mutation was successful, or an error message else.
        """
        license_str = None if not license else json.dumps(license)
        variables = {
            'id': organization_id
        }
        if name is not None:
            variables['name'] = name
        if address is not None:
            variables['address'] = address
        if license_str is not None:
            variables['license'] = license_str
        if zip_code is not None:
            variables['zipCode'] = zip_code
        if city is not None:
            variables['city'] = city
        if country is not None:
            variables['country'] = country
        result = self.auth.client.execute(
            GQL_UPDATE_PROPERTIES_IN_ORGANIZATION, variables)
        return format_result('data', result)
